import asyncio
import logging

from ofrak.component.unpacker import UnpackerError
from ofrak.model.viewable_tag_model import AttributesType
from ofrak_type.architecture import InstructionSetMode
from ofrak.core.addressable import Addressable
from ofrak.core.architecture import ProgramAttributes
from ofrak.core.basic_block import BasicBlock, BasicBlockUnpacker
from ofrak.core.instruction import (
    Instruction,
    InstructionAnalyzer,
    InstructionRegisterUsageAnalyzer,
    RegisterUsage,
)
from ofrak.resource import Resource, ResourceFactory
from ofrak.service.component_locator_i import ComponentLocatorInterface
from ofrak.service.data_service_i import DataServiceInterface
from ofrak.service.disassembler.disassembler_service_i import (
    DisassemblerServiceInterface,
    DisassemblerServiceRequest,
)
from ofrak.service.resource_service_i import ResourceServiceInterface

LOGGER = logging.getLogger(__name__)


class CapstoneBasicBlockUnpacker(BasicBlockUnpacker):
    def __init__(
        self,
        resource_factory: ResourceFactory,
        data_service: DataServiceInterface,
        resource_service: ResourceServiceInterface,
        component_locator: ComponentLocatorInterface,
        disassembler_service: DisassemblerServiceInterface,
    ):
        super().__init__(resource_factory, data_service, resource_service, component_locator)
        self._disassembler_service = disassembler_service

    async def unpack(self, resource: Resource, config=None):
        bb_view = await resource.view_as(BasicBlock)
        bb_data = await resource.get_data()

        program_attrs = await resource.analyze(ProgramAttributes)

        disassemble_request = DisassemblerServiceRequest(
            program_attrs.isa,
            program_attrs.sub_isa,
            program_attrs.bit_width,
            program_attrs.endianness,
            program_attrs.processor,
            bb_view.mode,
            bb_data,
            bb_view.virtual_address,
        )

        instruction_children_created = []

        for disassem_result in await self._disassembler_service.disassemble(disassemble_request):
            instruction_view = Instruction(
                disassem_result.address,
                disassem_result.size,
                f"{disassem_result.mnemonic} {disassem_result.operands}",
                disassem_result.mnemonic,
                disassem_result.operands,
                bb_view.mode,
            )

            instruction_children_created.append(
                bb_view.create_child_region(
                    instruction_view, additional_attributes=(program_attrs,)
                )
            )

        await asyncio.gather(*instruction_children_created)


class CapstoneInstructionAnalyzer(InstructionAnalyzer):
    def __init__(
        self,
        resource_factory: ResourceFactory,
        data_service: DataServiceInterface,
        resource_service: ResourceServiceInterface,
        disassembler_service: DisassemblerServiceInterface,
    ):
        super().__init__(resource_factory, data_service, resource_service)
        self._disassembler_service = disassembler_service

    async def analyze(self, resource: Resource, config=None) -> Instruction:
        parent_block = await resource.get_parent_as_view(Addressable)
        mode: InstructionSetMode
        if parent_block.resource.has_tag(BasicBlock):
            bb_attrs = await parent_block.resource.analyze(AttributesType[BasicBlock])
            mode = bb_attrs.mode  # type: ignore
        else:
            mode = InstructionSetMode.NONE
        instruction_data = await resource.get_data()
        program_attrs = await resource.analyze(ProgramAttributes)

        parent_start_vaddr = parent_block.virtual_address
        instr_offset_in_parent = (await resource.get_data_range_within_parent()).start

        disassemble_request = DisassemblerServiceRequest(
            program_attrs.isa,
            program_attrs.sub_isa,
            program_attrs.bit_width,
            program_attrs.endianness,
            program_attrs.processor,
            mode,
            instruction_data,
            parent_start_vaddr + instr_offset_in_parent,
        )

        try:
            disassem_result = next(
                iter(await self._disassembler_service.disassemble(disassemble_request))
            )
        except StopIteration:
            raise UnpackerError(
                f"Could not disassemble any {program_attrs.isa.name}-"
                f"{program_attrs.bit_width.name}-{program_attrs.endianness.name} instructions "
                f"from bytes {instruction_data.hex()}"
            )

        return Instruction(
            disassem_result.address,
            disassem_result.size,
            f"{disassem_result.mnemonic} {disassem_result.operands}",
            disassem_result.mnemonic,
            disassem_result.operands,
            mode,
        )


class CapstoneInstructionRegisterUsageAnalyzer(InstructionRegisterUsageAnalyzer):
    def __init__(
        self,
        resource_factory: ResourceFactory,
        data_service: DataServiceInterface,
        resource_service: ResourceServiceInterface,
        disassembler_service: DisassemblerServiceInterface,
    ):
        super().__init__(resource_factory, data_service, resource_service)
        self._disassembler_service = disassembler_service

    async def analyze(self, resource: Resource, config: None) -> RegisterUsage:
        program_attrs = await resource.analyze(ProgramAttributes)

        instruction = await resource.view_as(Instruction)

        disassemble_request = DisassemblerServiceRequest(
            program_attrs.isa,
            program_attrs.sub_isa,
            program_attrs.bit_width,
            program_attrs.endianness,
            program_attrs.processor,
            instruction.mode,
            await instruction.resource.get_data(),
            instruction.virtual_address,
        )

        disassem_result = await self._disassembler_service.get_register_usage(disassemble_request)

        return RegisterUsage(
            disassem_result.regs_read,
            disassem_result.regs_written,
        )



import logging
import re
from dataclasses import dataclass
from typing import Dict, Tuple, Optional, Iterable

from capstone import (
    Cs,
    CS_ARCH_ARM64,
    CS_ARCH_ARM,
    CS_ARCH_X86,
    CS_ARCH_PPC,
    CS_ARCH_MIPS,
    CS_MODE_BIG_ENDIAN,
    CS_MODE_LITTLE_ENDIAN,
    CS_MODE_THUMB,
    CS_MODE_ARM,
    CS_MODE_32,
    CS_MODE_64,
    CS_MODE_16,
    CsError,
)

from ofrak_type.architecture import (
    InstructionSet,
    SubInstructionSet,
    InstructionSetMode,
    ProcessorType,
)
from ofrak.service.disassembler.disassembler_service_i import (
    DisassemblerServiceInterface,
    DisassemblerServiceRequest,
    DisassemblyResult,
    RegisterUsageResult,
    DisassemblerRegisterUsageSupportError,
    DisassemblerArchSupportError,
)
from ofrak_type.bit_width import BitWidth
from ofrak_type.endianness import Endianness


@dataclass(frozen=True)
class CapstoneDisassemblerType:
    isa: InstructionSet
    sub_isa: Optional[SubInstructionSet]
    bit_width: BitWidth
    endianness: Endianness
    processor: Optional[ProcessorType]
    mode: InstructionSetMode


LOGGER = logging.getLogger(__file__)


RE_REPRESENT_CONSTANTS_HEX = re.compile(r"(\W-?)([0-9]([^0-9x]|$))")
RE_RENAME_FP_TO_R11 = re.compile(r"(\W?)fp(\W?)")


class CapstoneDisassemblerService(DisassemblerServiceInterface):
    def __init__(self):
        self._cs_instance_by_attributes: Dict[CapstoneDisassemblerType, Cs] = dict()
        self._cs_instances_with_no_register_access = set()

    def _cs_disassemble(self, request: DisassemblerServiceRequest):
        cs_disassembler_type = _get_cs_disam_type(request)
        cs = self._get_cs(cs_disassembler_type)

        return cs.disasm(request.data, request.virtual_address)

    async def disassemble(self, request: DisassemblerServiceRequest) -> Iterable[DisassemblyResult]:
        res = []

        for cs_instruction in self._cs_disassemble(request):
            mnemonic, operands = _asm_fixups(
                cs_instruction.mnemonic, cs_instruction.op_str, request.isa
            )

            res.append(
                DisassemblyResult(
                    cs_instruction.address,
                    cs_instruction.size,
                    mnemonic,
                    operands,
                )
            )
        return res

    async def get_register_usage(self, request: DisassemblerServiceRequest) -> RegisterUsageResult:
        for cs_instruction in self._cs_disassemble(request):
            if request.isa is InstructionSet.X86 and cs_instruction.mnemonic == "cmp":
                raise DisassemblerRegisterUsageSupportError(
                    f"Capstone's support for x86 'cmp' instructions is flaky and platform "
                    f"dependent! For consistency, a hard error is always raised."
                )
            try:
                (regs_read_indexes, regs_write_indexes) = cs_instruction.regs_access()
            except CsError:
                raise DisassemblerRegisterUsageSupportError(
                    f"Capstone cannot get register usages for {_get_cs_disam_type(request)}"
                )
            regs_read = tuple(cs_instruction.reg_name(i) for i in regs_read_indexes)
            regs_written = tuple(cs_instruction.reg_name(i) for i in regs_write_indexes)

            return RegisterUsageResult(regs_read, regs_written)

        raise ValueError(
            "Capstone could not get register usage info because it could not find and "
            "instructions in the provided data!"
        )

    def _get_cs(self, cs_disam: CapstoneDisassemblerType):
        cs = self._cs_instance_by_attributes.get(cs_disam)
        if cs is None:
            cs = Cs(
                self._get_cs_arch_flag(cs_disam),
                self._get_cs_mode_flag(cs_disam),
            )
            cs.detail = True
            self._cs_instance_by_attributes[cs_disam] = cs

        return cs

    @staticmethod
    def _get_cs_arch_flag(cs_disam: CapstoneDisassemblerType):
        isa = cs_disam.isa
        bit_width = cs_disam.bit_width

        if isa is InstructionSet.ARM:
            return CS_ARCH_ARM64 if bit_width is BitWidth.BIT_64 else CS_ARCH_ARM
        elif isa is InstructionSet.X86:
            return CS_ARCH_X86
        elif isa is InstructionSet.PPC:
            return CS_ARCH_PPC
        elif isa is InstructionSet.MIPS:
            return CS_ARCH_MIPS
        elif isa is InstructionSet.AARCH64:
            return CS_ARCH_ARM64
        raise DisassemblerArchSupportError(
            f"Cannot generate the capstone architecture flag for {cs_disam}"
        )

    @staticmethod
    def _get_cs_mode_flag(cs_disam: CapstoneDisassemblerType):
        isa = cs_disam.isa
        bit_width = cs_disam.bit_width
        endianness = cs_disam.endianness
        processor = cs_disam.processor
        mode = cs_disam.mode

        if endianness is Endianness.BIG_ENDIAN:
            cs_endian_flag = CS_MODE_BIG_ENDIAN
        else:
            cs_endian_flag = CS_MODE_LITTLE_ENDIAN

        if isa is InstructionSet.AARCH64:
            mode_flag = 0
            return mode_flag | cs_endian_flag

        if isa is InstructionSet.ARM:
            mode_flag = 0
            if mode is InstructionSetMode.THUMB:
                # THUMB mode
                return mode_flag | CS_MODE_THUMB | cs_endian_flag
            elif bit_width is BitWidth.BIT_32:
                mode_flag |= CS_MODE_ARM
            if processor is None:
                # Plain ARM 32
                return mode_flag | cs_endian_flag
            if processor is ProcessorType.XSCALE:
                # XSCALE
                raise NotImplementedError("XSCALE not implemented")
            elif processor is ProcessorType.ARM926EJ_S:
                # ARMv5
                raise NotImplementedError("ARMv5E not implemented")
            else:
                # Default 32-bit ARM for other Processor types
                return mode_flag | cs_endian_flag

        elif isa is InstructionSet.X86:
            if bit_width is BitWidth.BIT_64:
                return CS_MODE_64 | cs_endian_flag
            elif bit_width is BitWidth.BIT_32:
                return CS_MODE_32 | cs_endian_flag
            elif bit_width is BitWidth.BIT_16:
                return CS_MODE_16 | cs_endian_flag

        elif isa is InstructionSet.PPC:
            if bit_width == BitWidth.BIT_64:
                return CS_MODE_64 | cs_endian_flag
            elif bit_width == BitWidth.BIT_32:
                return CS_MODE_32 | cs_endian_flag

        elif isa is InstructionSet.MIPS:
            if bit_width == BitWidth.BIT_64:
                return CS_MODE_64 | cs_endian_flag
            elif bit_width == BitWidth.BIT_32:
                return CS_MODE_32 | cs_endian_flag

        raise DisassemblerArchSupportError(f"Cannot generate the capstone mode flag for {cs_disam}")


def _asm_fixups(base_mnemonic: str, base_operands: str, isa: InstructionSet) -> Tuple[str, str]:
    operands = re.sub(RE_REPRESENT_CONSTANTS_HEX, r"\g<1>0x\g<2>", base_operands)
    if isa is InstructionSet.ARM:
        operands = re.sub(RE_RENAME_FP_TO_R11, r"\1r11\2", operands)

    mnemonic = base_mnemonic

    return mnemonic, operands


def _get_cs_disam_type(request: DisassemblerServiceRequest) -> CapstoneDisassemblerType:
    return CapstoneDisassemblerType(
        request.isa,
        request.sub_isa,
        request.bit_width,
        request.endianness,
        request.processor,
        request.mode,
    )



import os
from binascii import unhexlify
from dataclasses import dataclass
from typing import Dict, List

import pytest

from ofrak.core import BinaryPatchModifier, BinaryPatchConfig
from ofrak.core.filesystem import File

import ofrak_capstone
from ofrak import OFRAKContext
from ofrak.model.viewable_tag_model import ResourceViewContext
from ofrak_type.architecture import InstructionSet, InstructionSetMode
from ofrak.core.architecture import ProgramAttributes
from ofrak.core.basic_block import BasicBlock
from ofrak.core.instruction import Instruction
from ofrak.core.program import Program
from ofrak.resource import Resource
from ofrak.service.resource_service_i import ResourceFilter, ResourceSort
from pytest_ofrak.patterns import TEST_PATTERN_ASSETS_DIR
from pytest_ofrak.patterns.basic_block_unpacker import (
    BasicBlockUnpackerUnpackAndVerifyPattern,
    BasicBlockUnpackerTestCase,
)
from pytest_ofrak.patterns.register_usage_analyzer import (
    RegisterAnalyzerTestCase,
    RegisterUsageTestPattern,
)
from test_ofrak.constants import ARM32_ARCH

pytest_plugins = ["pytest_ofrak.fixtures"]


@pytest.fixture
def test_id():
    return "TEST_JOB"


@pytest.fixture(autouse=True)
def capstone_components(ofrak_injector):
    ofrak_injector.discover(ofrak_capstone)


class TestCapstoneBasicBlockUnpackAndVerify(BasicBlockUnpackerUnpackAndVerifyPattern):
    async def unpack(self, root_resource: Resource):
        basic_block_resources = await root_resource.get_descendants(
            r_filter=ResourceFilter.with_tags(BasicBlock),
        )
        for bb_r in basic_block_resources:
            await bb_r.unpack()

    @pytest.fixture
    async def root_resource(
        self,
        unpack_verify_test_case: BasicBlockUnpackerTestCase,
        ofrak_context: OFRAKContext,
        test_id: str,
    ) -> Resource:
        """
        Some hacky logic in here to avoid needing to unpack the whole resource top to bottom
        (which would require an analysis backend) and instead define the basic blocks manually.
        The hacky part is that the root resource is duplicated with all its data, once as the
        root resource (which gets unpacked so that we can analyze the ProgramAttributes for the
        capstone BB unpacker's benefit) and a second time as an unmapped child of that root
        resource, which does not get unpacked but instead has basic block children manually
        mapped into it.
        """
        asset_path = os.path.join(TEST_PATTERN_ASSETS_DIR, unpack_verify_test_case.binary_filename)
        with open(asset_path, "rb") as f:
            binary_data = f.read()
        top_resource = await ofrak_context.create_root_resource(test_id, binary_data, tags=(File,))
        await top_resource.unpack()
        resource = await top_resource.create_child(
            data=binary_data,
        )

        for bb_vaddr, bb_range in unpack_verify_test_case.basic_block_data_ranges_in_root.items():
            await resource.create_child_from_view(
                BasicBlock(bb_vaddr, bb_range.length(), InstructionSetMode.NONE, False, None),
                data_range=bb_range,
            )
        return resource

    async def get_descendants_to_verify(self, unpacked_resource: Resource) -> Dict[int, Resource]:
        basic_blocks = await unpacked_resource.get_descendants_as_view(
            BasicBlock,
            r_filter=ResourceFilter.with_tags(BasicBlock),
            r_sort=ResourceSort(BasicBlock.VirtualAddress),
        )
        return {bb.virtual_address: bb for bb in basic_blocks}


@dataclass
class UnpackerTestCase:
    label: str
    program_attributes: ProgramAttributes
    parent_basic_block: BasicBlock
    basic_block_data: bytes
    expected_children: List[Instruction]

    async def run_instruction_unpacker_test_case(self, ofrak: OFRAKContext):
        bb_r = await self._initialize_basic_bloc_resource(ofrak)

        await bb_r.unpack()

        children = list(
            await bb_r.get_children_as_view(
                Instruction, r_sort=ResourceSort(BasicBlock.VirtualAddress)
            )
        )

        assert len(self.expected_children) == len(children), (
            f"Did not unpack the expected number"
            f" of children (expected {len(self.expected_children)})"
        )

        for expected_child, child in zip(self.expected_children, children):
            assert expected_child == child

    async def run_instruction_analzyer_test_case(self, ofrak: OFRAKContext):
        bb_r = await self._initialize_basic_bloc_resource(ofrak)
        await bb_r.unpack()
        basic_block = await bb_r.view_as(BasicBlock)
        for instruction in await basic_block.get_instructions():
            raw_bytes = await instruction.resource.get_data()
            # Patch ourselves
            await instruction.resource.run(BinaryPatchModifier, BinaryPatchConfig(0, raw_bytes))
            # Clear context to get resource view again, triggering InstructionAnalyzer. Without
            # this step the InstructionAnalyzer does not run.
            instruction.resource._resource_view_context = ResourceViewContext()
            reanalyzed_instruction = await instruction.resource.view_as(Instruction)
            assert instruction == reanalyzed_instruction

    async def _initialize_basic_bloc_resource(self, ofrak) -> Resource:
        bb_r = await ofrak.create_root_resource(
            "test_resource", self.basic_block_data, tags=(Program,)
        )
        bb_r.add_view(self.parent_basic_block)
        bb_r.add_attributes(self.program_attributes)
        await bb_r.save()
        return bb_r


BASIC_BLOCK_TEST_CASES = [
    UnpackerTestCase(
        "Arm32",
        ARM32_ARCH,
        BasicBlock(
            0x100,
            0x14,
            InstructionSetMode.NONE,
            False,
            None,
        ),
        unhexlify("F0412DE90040A0E30450A0E150609FE550709FE5"),
        [
            Instruction(
                0x100,
                0x4,
                "push {r4, r5, r6, r7, r8, lr}",
                "push",
                "{r4, r5, r6, r7, r8, lr}",
                InstructionSetMode.NONE,
            ),
            Instruction(
                0x104,
                0x4,
                "mov r4, #0x0",
                "mov",
                "r4, #0x0",
                InstructionSetMode.NONE,
            ),
            Instruction(
                0x108,
                0x4,
                "mov r5, r4",
                "mov",
                "r5, r4",
                InstructionSetMode.NONE,
            ),
            Instruction(
                0x10C,
                0x4,
                "ldr r6, [pc, #0x50]",
                "ldr",
                "r6, [pc, #0x50]",
                InstructionSetMode.NONE,
            ),
            Instruction(
                0x110,
                0x4,
                "ldr r7, [pc, #0x50]",
                "ldr",
                "r7, [pc, #0x50]",
                InstructionSetMode.NONE,
            ),
        ],
    )
]


@pytest.mark.parametrize("test_case", BASIC_BLOCK_TEST_CASES, ids=lambda tc: tc.label)
async def test_capstone_unpacker(test_case, ofrak_context):
    await test_case.run_instruction_analzyer_test_case(ofrak_context)


@pytest.mark.parametrize("test_case", BASIC_BLOCK_TEST_CASES, ids=lambda tc: tc.label)
async def test_capstone_analyzer(test_case, ofrak_context):
    await test_case.run_instruction_analzyer_test_case(ofrak_context)


class TestCapstoneRegisterUsage(RegisterUsageTestPattern):
    def case_is_known_broken(self, test_case: RegisterAnalyzerTestCase):
        if test_case.program_attributes.isa is InstructionSet.PPC:
            return True, "capstone fails to give register usage info for PPC instructions"

        elif test_case.program_attributes.isa is InstructionSet.X86:
            mnemonics = {"call"}
            if (
                test_case.instruction.mnemonic not in mnemonics
                and "rip" not in test_case.instruction.operands
                and "rip" in test_case.expected_regs_read
            ):
                return (
                    True,
                    "capstone fails to find some implicit rip reads",
                )
        return False, ""



import setuptools
from setuptools.command.egg_info import egg_info


class egg_info_ex(egg_info):
    """Includes license file into `.egg-info` folder."""

    def run(self):
        # don't duplicate license into `.egg-info` when building a distribution
        if not self.distribution.have_run.get("install", True):
            # `install` command is in progress, copy license
            self.mkpath(self.egg_info)
            self.copy_file("LICENSE", self.egg_info)

        egg_info.run(self)


with open("README.md") as f:
    long_description = f.read()

setuptools.setup(
    name="ofrak_capstone",
    version="1.0.0",
    packages=setuptools.find_packages(exclude=["ofrak_capstone_test", "ofrak_capstone_test.*"]),
    package_data={"ofrak_capstone": ["py.typed"]},
    install_requires=["capstone==4.0.2", "ofrak"],
    extras_require={
        "test": [
            "fun-coverage==0.2.0",
            "pytest",
            "pytest-cov",
            "pytest-asyncio==0.19.0",
            "requests",
            "ofrak[test]",
        ]
    },
    author="Red Balloon Security",
    author_email="ofrak@redballoonsecurity.com",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://ofrak.com/",
    download_url="https://github.com/redballoonsecurity/ofrak",
    project_urls={
        "Documentation": "https://ofrak.com/docs/",
        "Community License": "https://github.com/redballoonsecurity/ofrak/blob/master/LICENSE",
        "Commercial Licensing Information": "https://ofrak.com/license/",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: Other/Proprietary License",
        "License :: Free To Use But Restricted",
        "License :: Free For Home Use",
        "Topic :: Security",
        "Typing :: Typed",
    ],
    python_requires=">=3.7",
    license="Proprietary",
    license_files=["LICENSE"],
    cmdclass={"egg_info": egg_info_ex},
    entry_points={"ofrak.packages": ["ofrak_capstone_pkg = ofrak_capstone"]},
)
